#pragma once
//
// AUTO-GENERATED at CMake configure time from LEMON_BACKENDS in CMakeLists.txt.
// Do not edit by hand. Descriptor DATA only (CLI-safe; no server classes).
//
#include <vector>
#include "lemon/backends/backend_descriptor.h"
#include "lemon/backends/llamacpp/llamacpp.h"
#include "lemon/backends/whispercpp/whispercpp.h"
#include "lemon/backends/moonshine/moonshine.h"
#include "lemon/backends/kokoro/kokoro.h"
#include "lemon/backends/sdcpp/sdcpp.h"
#include "lemon/backends/fastflowlm/fastflowlm.h"
#include "lemon/backends/ryzenai/ryzenai.h"
#include "lemon/backends/vllm/vllm.h"
#include "lemon/backends/cloud/cloud.h"
#include "lemon/backends/thinksound/thinksound.h"
#include "lemon/backends/acestep/acestep.h"
#include "lemon/backends/onnxruntime/onnxruntime.h"
#include "lemon/backends/trellis/trellis.h"
#include "lemon/backends/openmoss/openmoss.h"

namespace lemon {
namespace backends {

inline std::vector<const BackendDescriptor*> all_generated_descriptors() {
    return {
        &lemon::backends::llamacpp::descriptor,
        &lemon::backends::whispercpp::descriptor,
        &lemon::backends::moonshine::descriptor,
        &lemon::backends::kokoro::descriptor,
        &lemon::backends::sdcpp::descriptor,
        &lemon::backends::fastflowlm::descriptor,
        &lemon::backends::ryzenai::descriptor,
        &lemon::backends::vllm::descriptor,
        &lemon::backends::cloud::descriptor,
        &lemon::backends::thinksound::descriptor,
        &lemon::backends::acestep::descriptor,
        &lemon::backends::onnxruntime::descriptor,
        &lemon::backends::trellis::descriptor,
        &lemon::backends::openmoss::descriptor,

    };
}

} // namespace backends
} // namespace lemon
